import asyncio
import base64
from email.message import EmailMessage
import uuid

import pytest

from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.mail_channel import MailChannel
from meshagent.agents.mail_common import SmtpConfiguration
from meshagent.agents.messages import (
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    TurnEnded,
    TurnStart,
    TurnStarted,
)
from meshagent.agents.process import Message
from meshagent.agents.thread_schema import thread_list_schema
from meshagent.api import Participant
from meshagent.api.messaging import FileContent
from meshagent.api.room_server_client import RoomException
from meshagent.tools import ToolContext


class _FakeLocalParticipant(Participant):
    def __init__(self) -> None:
        super().__init__(id="assistant-id", attributes={"name": "assistant"})
        self.set_attribute_calls: list[tuple[str, object]] = []

    async def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value
        self.set_attribute_calls.append((name, value))


class _FakeProtocol:
    def __init__(self) -> None:
        self.token = "token"


class _FakeQueues:
    def __init__(self) -> None:
        self._queue: asyncio.Queue[dict] = asyncio.Queue()
        self.receive_calls: list[dict[str, object]] = []

    async def receive(self, *, name: str, create: bool, wait: bool):
        self.receive_calls.append({"name": name, "create": create, "wait": wait})
        return await self._queue.get()

    async def push(self, payload: dict) -> None:
        await self._queue.put(payload)


class _FakeStorage:
    def __init__(self) -> None:
        self.uploaded: dict[str, bytes] = {}
        self.exists_calls: list[str] = []

    async def upload(self, *, path: str, data: bytes) -> None:
        self.uploaded[path] = data

    async def exists(self, *, path: str) -> bool:
        self.exists_calls.append(path)
        return path in self.uploaded

    async def download(self, *, path: str) -> FileContent:
        return FileContent(
            data=self.uploaded[path],
            name=path.rsplit("/", 1)[-1],
            mime_type="application/octet-stream",
        )


class _FakeDatabase:
    def __init__(self) -> None:
        self.tables: dict[tuple[tuple[str, ...], str], list[dict[str, object]]] = {}
        self.create_calls: list[dict[str, object]] = []

    def _key(
        self, *, table: str, namespace: list[str] | None
    ) -> tuple[tuple[str, ...], str]:
        return (tuple(namespace or []), table)

    async def create_table_with_schema(
        self,
        *,
        name: str,
        schema,
        mode: str,
        namespace: list[str] | None = None,
    ) -> None:
        del schema
        self.create_calls.append(
            {"name": name, "mode": mode, "namespace": list(namespace or [])}
        )
        self.tables.setdefault(self._key(table=name, namespace=namespace), [])

    async def insert(
        self,
        *,
        table: str,
        records: list[dict[str, object]],
        namespace: list[str] | None = None,
    ) -> None:
        self.tables.setdefault(self._key(table=table, namespace=namespace), []).extend(
            [dict(record) for record in records]
        )

    async def search(
        self,
        *,
        table: str,
        where,
        namespace: list[str] | None = None,
    ) -> list[dict[str, object]]:
        rows = self.tables.setdefault(self._key(table=table, namespace=namespace), [])
        if isinstance(where, dict):
            return [
                row
                for row in rows
                if all(row.get(key) == value for key, value in where.items())
            ]
        raise AssertionError("test database only supports dict where clauses")


class _NamespaceRejectingDatabase(_FakeDatabase):
    async def create_table_with_schema(
        self,
        *,
        name: str,
        schema,
        mode: str,
        namespace: list[str] | None = None,
    ) -> None:
        self.create_calls.append(
            {"name": name, "mode": mode, "namespace": list(namespace or [])}
        )
        if namespace:
            raise RoomException(
                "Error creating table 'emails': Invalid input, Location must be provided when namespace is not empty"
            )
        del schema
        self.tables.setdefault(self._key(table=name, namespace=namespace), [])


class _FakeThreadListElement:
    def __init__(self, *, tag_name: str, attributes: dict[str, str]) -> None:
        self.tag_name = tag_name
        self._attributes = dict(attributes)

    def get_attribute(self, name: str):
        return self._attributes.get(name)

    def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value


class _FakeThreadListRoot:
    def __init__(self) -> None:
        self._children: list[_FakeThreadListElement] = []

    def get_children(self) -> list[_FakeThreadListElement]:
        return [*self._children]

    def append_child(
        self,
        *,
        tag_name: str,
        attributes: dict[str, str],
    ) -> _FakeThreadListElement:
        element = _FakeThreadListElement(tag_name=tag_name, attributes=attributes)
        self._children.append(element)
        return element


class _FakeThreadListDocument:
    def __init__(self) -> None:
        self.root = _FakeThreadListRoot()


class _FakeSync:
    def __init__(self) -> None:
        self.document = _FakeThreadListDocument()
        self.open_calls: list[dict[str, object]] = []
        self.close_calls: list[str] = []

    async def open(self, *, path: str, schema=None) -> _FakeThreadListDocument:
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str) -> None:
        self.close_calls.append(path)


class _FakeRoom:
    def __init__(self, *, database: _FakeDatabase | None = None) -> None:
        self.local_participant = _FakeLocalParticipant()
        self.protocol = _FakeProtocol()
        self.queues = _FakeQueues()
        self.storage = _FakeStorage()
        self.database = database if database is not None else _FakeDatabase()
        self.sync = _FakeSync()


class _RecordingSupervisor:
    def __init__(self) -> None:
        self.sent: list[Message] = []

    def send(self, message: Message) -> None:
        self.sent.append(message)


class _FakeThreadNameAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str) -> None:
        self.generated_thread_name = generated_thread_name
        self.prompts: list[str] = []

    def default_model(self) -> str:
        return "thread-name-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options=None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        self.prompts = [
            message["content"]
            for message in context.messages
            if isinstance(message, dict) and isinstance(message.get("content"), str)
        ]
        return {"thread_name": self.generated_thread_name}


def _assert_uuid_thread_path(*, path: str, prefix: str) -> None:
    assert path.startswith(prefix)
    assert path.endswith(".thread")
    basename = path[len(prefix) : -len(".thread")]
    parsed = uuid.UUID(basename)
    assert str(parsed) == basename


def _email_bytes(
    *,
    from_address: str,
    to_address: str,
    subject: str,
    body: str,
    in_reply_to: str | None = None,
    attachments: list[tuple[str, bytes]] | None = None,
) -> bytes:
    message = EmailMessage()
    message["From"] = from_address
    message["To"] = to_address
    message["Subject"] = subject
    if in_reply_to is not None:
        message["In-Reply-To"] = in_reply_to
    message.set_content(body)
    for file_name, data in attachments or []:
        message.add_attachment(
            data,
            maintype="application",
            subtype="octet-stream",
            filename=file_name,
        )
    return message.as_bytes()


async def _drain() -> None:
    await asyncio.sleep(0)
    await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_mail_channel_creates_new_thread_and_persists_thread_mapping() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
    )
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        raw_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Quarterly update",
            body="hello from email",
            attachments=[("report.txt", b"hi")],
        )
        await room.queues.push(
            {"base64": base64.b64encode(raw_message).decode("ascii")}
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        _assert_uuid_thread_path(
            path=outbound.data.thread_id,
            prefix=".threads/assistant/",
        )
        assert outbound.sender is not None
        assert outbound.sender.get_attribute("name") == "Alice"
        assert room.database.create_calls == [
            {
                "name": "emails",
                "mode": "create_if_not_exists",
                "namespace": [".threads", "assistant"],
            }
        ]
        stored_rows = room.database.tables[(((".threads", "assistant")), "emails")]
        assert len(stored_rows) == 1
        assert stored_rows[0]["thread_id"] == outbound.data.thread_id
        assert any(
            path.endswith("/attachments/report.txt") for path in room.storage.uploaded
        )
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_maps_reply_to_existing_thread(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        thread_dir=".threads/helpdesk",
    )

    async def fake_send(message, *, hostname, port, username, password, local_hostname):
        del message
        del hostname
        del port
        del username
        del password
        del local_hostname

    monkeypatch.setattr("meshagent.agents.mail_channel.aiosmtplib.send", fake_send)
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        original_saved = await channel.start_thread(
            thread_id=".threads/helpdesk/ticket.thread",
            to_address="bob@example.com",
            subject="Initial",
            body="hi",
        )
        reply = _email_bytes(
            from_address="Bob <bob@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Re: Initial",
            body="following up",
            in_reply_to=original_saved["id"],
        )
        await room.queues.push({"base64": base64.b64encode(reply).decode("ascii")})
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.thread_id == ".threads/helpdesk/ticket.thread"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_default_new_indexes_new_threads_from_subject() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        threading_mode="default-new",
    )
    expected_uuid = uuid.UUID("12345678-1234-5678-1234-567812345678")

    original_uuid4 = uuid.uuid4
    uuid.uuid4 = lambda: expected_uuid
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        raw_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Quarterly update",
            body="hello from email",
        )
        await room.queues.push(
            {"base64": base64.b64encode(raw_message).decode("ascii")}
        )
        await _drain()

        assert room.local_participant.set_attribute_calls == [
            ("meshagent.chatbot.threading", "default-new"),
            ("meshagent.chatbot.thread-dir", ".threads/assistant"),
            ("meshagent.chatbot.thread-list", ".threads/assistant/index.threadl"),
        ]
        assert room.sync.open_calls == [
            {
                "path": ".threads/assistant/index.threadl",
                "schema": thread_list_schema,
            }
        ]
        assert room.storage.exists_calls == [
            ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        ]

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert (
            outbound.data.thread_id
            == ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        )

        entries = room.sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Quarterly update"
        assert (
            entries[0].get_attribute("path")
            == ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        )
    finally:
        uuid.uuid4 = original_uuid4
        await channel.stop(supervisor)  # type: ignore[arg-type]

    assert room.sync.close_calls == [".threads/assistant/index.threadl"]


@pytest.mark.asyncio
async def test_mail_channel_default_new_uses_attachment_names_for_llm_thread_naming() -> (
    None
):
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    adapter = _FakeThreadNameAdapter(generated_thread_name="Quarterly Files")
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        threading_mode="default-new",
        llm_adapter=adapter,
    )

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        raw_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="",
            body="See the files",
            attachments=[("quarterly-report.pdf", b"pdf-bytes")],
        )
        await room.queues.push(
            {"base64": base64.b64encode(raw_message).decode("ascii")}
        )
        await _drain()

        assert adapter.prompts == [
            "Message:\nSee the files\n\nAttachments:\n- quarterly-report.pdf"
        ]
        entries = room.sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Quarterly Files"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_default_new_reply_keeps_existing_thread_list_name() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        threading_mode="default-new",
    )

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        original_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Quarterly update",
            body="hello from email",
        )
        await room.queues.push(
            {"base64": base64.b64encode(original_message).decode("ascii")}
        )
        await _drain()

        stored_rows = room.database.tables[(((".threads", "assistant")), "emails")]
        reply_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Re: Quarterly update",
            body="following up",
            in_reply_to=stored_rows[0]["id"],
        )
        await room.queues.push(
            {"base64": base64.b64encode(reply_message).decode("ascii")}
        )
        await _drain()

        entries = room.sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Quarterly update"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_falls_back_to_flat_table_when_namespace_create_is_rejected() -> (
    None
):
    database = _NamespaceRejectingDatabase()
    room = _FakeRoom(database=database)
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        thread_dir=".threads/helpdesk",
    )
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        raw_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Fallback path",
            body="hello from email",
        )
        await room.queues.push(
            {"base64": base64.b64encode(raw_message).decode("ascii")}
        )
        await _drain()

        assert database.create_calls == [
            {
                "name": "emails",
                "mode": "create_if_not_exists",
                "namespace": [".threads", "helpdesk"],
            },
            {
                "name": "emails__threads__helpdesk",
                "mode": "create_if_not_exists",
                "namespace": [],
            },
        ]
        stored_rows = database.tables[((), "emails__threads__helpdesk")]
        assert len(stored_rows) == 1
        thread_id = stored_rows[0]["thread_id"]
        assert isinstance(thread_id, str)
        _assert_uuid_thread_path(path=thread_id, prefix=".threads/helpdesk/")
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_sends_reply_when_turn_ends(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
    )
    sent_messages: list[EmailMessage] = []

    async def fake_send(message, *, hostname, port, username, password, local_hostname):
        del hostname
        del port
        del username
        del password
        del local_hostname
        sent_messages.append(message)

    monkeypatch.setattr("meshagent.agents.mail_channel.aiosmtplib.send", fake_send)

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        raw_message = _email_bytes(
            from_address="Alice <alice@example.com>",
            to_address="mailbox@mail.meshagent.com",
            subject="Need help",
            body="please respond",
        )
        await room.queues.push(
            {"base64": base64.b64encode(raw_message).decode("ascii")}
        )
        await _drain()

        turn_start = supervisor.sent[0].data
        assert isinstance(turn_start, TurnStart)

        channel.send(
            Message(
                data=TurnStarted(
                    type="meshagent.agent.turn.started",
                    thread_id=turn_start.thread_id,
                    turn_id="turn-1",
                    source_message_id=turn_start.message_id,
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentStarted(
                    type="meshagent.agent.text_content.started",
                    thread_id=turn_start.thread_id,
                    turn_id="turn-1",
                    item_id="item-1",
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentDelta(
                    type="meshagent.agent.text_content.delta",
                    thread_id=turn_start.thread_id,
                    turn_id="turn-1",
                    item_id="item-1",
                    text="Thanks for the note.",
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentEnded(
                    type="meshagent.agent.text_content.ended",
                    thread_id=turn_start.thread_id,
                    turn_id="turn-1",
                    item_id="item-1",
                )
            )
        )
        channel.send(
            Message(
                data=TurnEnded(
                    type="meshagent.agent.turn.ended",
                    thread_id=turn_start.thread_id,
                    turn_id="turn-1",
                    error=None,
                )
            )
        )
        await _drain()

        assert len(sent_messages) == 1
        assert sent_messages[0]["Subject"] == "RE: Need help"
        assert (
            "Thanks for the note."
            in sent_messages[0].get_body(("plain",)).get_content()
        )
        stored_rows = room.database.tables[(((".threads", "assistant")), "emails")]
        assert len(stored_rows) == 2
        assert stored_rows[-1]["thread_id"] == turn_start.thread_id
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_new_email_thread_tool_uses_current_thread_id(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    room = _FakeRoom()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
    )
    sent_messages: list[EmailMessage] = []

    async def fake_send(message, *, hostname, port, username, password, local_hostname):
        del hostname
        del port
        del username
        del password
        del local_hostname
        sent_messages.append(message)

    monkeypatch.setattr("meshagent.agents.mail_channel.aiosmtplib.send", fake_send)

    await channel.start(_RecordingSupervisor())  # type: ignore[arg-type]
    try:
        room.storage.uploaded["notes.txt"] = b"hello"
        toolkit = channel.get_agent_toolkits()[0]
        new_thread_tool = toolkit.tools[0]
        await new_thread_tool.execute(
            ToolContext(
                room=room,  # type: ignore[arg-type]
                caller=room.local_participant,
                caller_context={"thread_id": ".threads/assistant/customer.thread"},
            ),
            to="customer@example.com",
            subject="Follow up",
            body="Here is the update.",
            attachments=["notes.txt"],
        )

        assert len(sent_messages) == 1
        assert sent_messages[0]["To"] == "customer@example.com"
        rows = room.database.tables[(((".threads", "assistant")), "emails")]
        assert len(rows) == 1
        assert rows[0]["thread_id"] == ".threads/assistant/customer.thread"
    finally:
        await channel.stop(channel.supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_mail_channel_uses_roompool_hostname_fallback_for_smtp(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.delenv("SMTP_USERNAME", raising=False)
    monkeypatch.delenv("SMTP_PASSWORD", raising=False)
    monkeypatch.delenv("SMTP_PORT", raising=False)
    monkeypatch.delenv("SMTP_HOSTNAME", raising=False)
    monkeypatch.delenv("SMTP_LOCAL_HOSTNAME", raising=False)
    monkeypatch.setenv("HOSTNAME", "roompool-rpt6h-r2677")
    monkeypatch.setattr("meshagent.agents.mail_common.socket.getfqdn", lambda: "")
    monkeypatch.setattr("meshagent.agents.mail_common.socket.gethostname", lambda: "")

    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = MailChannel(
        room=room,
        queue_name="mailbox@mail.meshagent.com",
        email_address="mailbox@mail.meshagent.com",
        domain="mail.meshagent.com",
        smtp=SmtpConfiguration(),
    )
    sent: dict[str, object] = {}

    async def fake_send(
        message,
        *,
        hostname,
        port,
        username,
        password,
        local_hostname,
    ):
        del message
        sent["hostname"] = hostname
        sent["port"] = port
        sent["username"] = username
        sent["password"] = password
        sent["local_hostname"] = local_hostname

    monkeypatch.setattr("meshagent.agents.mail_channel.aiosmtplib.send", fake_send)

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await channel.start_thread(
            thread_id=".threads/assistant/customer.thread",
            to_address="customer@example.com",
            subject="Follow up",
            body="Here is the update.",
        )
        assert sent == {
            "hostname": "mail.meshagent.com",
            "port": 587,
            "username": "assistant",
            "password": "token",
            "local_hostname": "roompool-rpt6h-r2677",
        }
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]



from typing import Optional

import pytest

from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.context import AgentSessionContext
from meshagent.agents.worker import Worker


class _FakeParticipant:
    def __init__(self, *, name: str, participant_id: str):
        self._name = name
        self.id = participant_id

    def get_attribute(self, key: str):
        if key == "name":
            return self._name
        return None


class _FakeElement:
    def __init__(self, *, tag_name: str, attributes: Optional[dict] = None):
        self.tag_name = tag_name
        self._attributes = dict(attributes or {})
        self._children: list["_FakeElement"] = []

    def get_attribute(self, key: str):
        return self._attributes.get(key)

    def set_attribute(self, key: str, value) -> None:
        self._attributes[key] = value

    def get_children(self):
        return self._children

    def append_child(self, tag_name: str, attributes: Optional[dict] = None):
        child = _FakeElement(tag_name=tag_name, attributes=attributes)
        self._children.append(child)
        return child


class _FakeThreadRoot:
    def __init__(self):
        self._members = _FakeElement(tag_name="members")

    def get_children_by_tag_name(self, tag_name: str):
        if tag_name == "members":
            return [self._members]
        return []

    @property
    def members(self) -> _FakeElement:
        return self._members


class _FakeThreadDocument:
    def __init__(self):
        self.root = _FakeThreadRoot()

    @property
    def member_names(self) -> list[str]:
        names: list[str] = []
        for child in self.root.members.get_children():
            if child.tag_name != "member":
                continue
            name = child.get_attribute("name")
            if isinstance(name, str):
                names.append(name)
        return names


class _FakeThreadListDocument:
    def __init__(self):
        self.root = _FakeElement(tag_name="thread_list")

    def get_state(self) -> bytes:
        return b"thread-list-state"


class _FakeSync:
    def __init__(self):
        self.document = _FakeThreadListDocument()
        self.open_calls: list[dict] = []
        self.close_calls: list[str] = []
        self.sync_calls: list[dict] = []

    async def open(self, *, path: str, schema=None):
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str):
        self.close_calls.append(path)

    async def sync(self, *, path: str, data: bytes):
        self.sync_calls.append({"path": path, "data": data})


class _FakeStorage:
    def __init__(self, *, existing_paths: Optional[set[str]] = None):
        self._existing_paths = set(existing_paths or set())
        self.exists_calls: list[str] = []

    async def exists(self, *, path: str) -> bool:
        self.exists_calls.append(path)
        return path in self._existing_paths


class _FakeRoom:
    def __init__(self, *, existing_paths: Optional[set[str]] = None):
        self.local_participant = _FakeParticipant(
            name="assistant",
            participant_id="assistant-id",
        )
        self.sync = _FakeSync()
        self.storage = _FakeStorage(existing_paths=existing_paths)


class _FakeThreadAdapter:
    instances: list["_FakeThreadAdapter"] = []

    def __init__(self, *, room, path: str):
        del room
        self.path = path
        self.thread = _FakeThreadDocument()
        self.started = False
        self.stopped = False
        self.appended = False
        self.writes: list[tuple[str, str]] = []
        self.events: list[dict] = []
        _FakeThreadAdapter.instances.append(self)

    async def start(self) -> None:
        self.started = True

    async def stop(self) -> None:
        self.stopped = True

    def append_messages(self, *, context: AgentSessionContext) -> None:
        del context
        self.appended = True

    def write_text_message(self, *, text: str, participant) -> None:
        self.writes.append((text, participant))

    def push(self, *, event: dict) -> None:
        self.events.append(event)


class _FakeLLMAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str = "release planning"):
        self.generated_thread_name = generated_thread_name
        self.calls: list[dict] = []

    def default_model(self) -> str:
        return "test-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del context, room, toolkits, steering_callback, on_behalf_of, options
        self.calls.append(
            {
                "output_schema": output_schema,
                "event_handler": event_handler,
                "model": model,
            }
        )

        if output_schema is not None:
            return {"thread_name": self.generated_thread_name}

        if event_handler is not None:
            event_handler({"type": "response.output_text.done", "text": "done"})

        return "assistant response"


class _FakeDecisionLLMAdapter(LLMAdapter):
    def __init__(self, *, summary: str):
        self._summary = summary
        self.calls: list[dict] = []

    def default_model(self) -> str:
        return "decision-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del (
            context,
            room,
            toolkits,
            event_handler,
            steering_callback,
            on_behalf_of,
            options,
        )
        self.calls.append({"output_schema": output_schema, "model": model})
        return {"summary": self._summary}


@pytest.mark.asyncio
async def test_worker_auto_threading_creates_thread_and_index_entry() -> None:
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter(generated_thread_name="Release Planning")
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="auto",
        thread_dir="/threads",
    )
    room = _FakeRoom()
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    context = AgentSessionContext(system_role=None)
    result = await worker.process_message(
        chat_context=context,
        message={"prompt": "Plan the release"},
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 2
    assert adapter.calls[0]["output_schema"] is not None
    assert adapter.calls[1]["output_schema"] is None
    assert adapter.calls[1]["event_handler"] is not None

    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.path == "/threads/release-planning.thread"
    assert thread_adapter.started
    assert thread_adapter.appended
    assert thread_adapter.stopped
    assert thread_adapter.thread.member_names == ["assistant"]
    assert thread_adapter.writes == [("```text\nPlan the release\n```", "worker")]
    assert [event["type"] for event in thread_adapter.events] == [
        "response.output_text.done"
    ]

    assert room.sync.open_calls[0]["path"] == "/threads/index.threadl"
    assert room.sync.sync_calls[0]["path"] == "/threads/index.threadl"
    assert room.sync.sync_calls[0]["data"] == b"dGhyZWFkLWxpc3Qtc3RhdGU="
    assert room.sync.close_calls == ["/threads/index.threadl"]
    entries = room.sync.document.root.get_children()
    assert len(entries) == 1
    assert entries[0].get_attribute("path") == "/threads/release-planning.thread"


@pytest.mark.asyncio
async def test_worker_auto_threading_uses_next_available_path_when_base_exists():
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter(generated_thread_name="Release Planning")
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="auto",
        thread_dir="/threads",
    )
    room = _FakeRoom(existing_paths={"/threads/release-planning.thread"})
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    result = await worker.process_message(
        chat_context=AgentSessionContext(system_role=None),
        message={"prompt": "Plan the release"},
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.path == "/threads/release-planning 2.thread"
    assert room.sync.open_calls[0]["path"] == "/threads/index.threadl"
    entries = room.sync.document.root.get_children()
    assert len(entries) == 1
    assert entries[0].get_attribute("path") == "/threads/release-planning 2.thread"


@pytest.mark.asyncio
async def test_worker_manual_threading_uses_message_path_without_index() -> None:
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter()
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="manual",
        thread_dir="/threads",
    )
    room = _FakeRoom()
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    result = await worker.process_message(
        chat_context=AgentSessionContext(system_role=None),
        message={"prompt": "Manual task", "path": "/threads/manual.thread"},
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 1
    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.path == "/threads/manual.thread"
    assert thread_adapter.thread.member_names == ["assistant"]
    assert thread_adapter.writes == [("```text\nManual task\n```", "worker")]
    assert room.sync.open_calls == []
    assert room.sync.close_calls == []


@pytest.mark.asyncio
async def test_worker_initial_message_summary_mode_uses_decision_adapter() -> None:
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter()
    decision_adapter = _FakeDecisionLLMAdapter(summary="Incoming webhook event summary")
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="manual",
        thread_dir="/threads",
        initial_message_mode="summary",
        initial_message_from="Webhook",
        decision_model="gpt-5.2-mini",
        decision_llm_adapter=decision_adapter,
    )
    room = _FakeRoom()
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    result = await worker.process_message(
        chat_context=AgentSessionContext(system_role=None),
        message={
            "prompt": "Process webhook payload",
            "path": "/threads/manual.thread",
        },
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(decision_adapter.calls) == 1
    assert decision_adapter.calls[0]["output_schema"] is not None
    assert decision_adapter.calls[0]["model"] == "gpt-5.2-mini"
    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.writes == [("Incoming webhook event summary", "Webhook")]


@pytest.mark.asyncio
async def test_worker_initial_message_none_mode_skips_thread_write() -> None:
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter()
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="manual",
        thread_dir="/threads",
        initial_message_mode="none",
    )
    room = _FakeRoom()
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    result = await worker.process_message(
        chat_context=AgentSessionContext(system_role=None),
        message={
            "prompt": "Manual task",
            "path": "/threads/manual.thread",
        },
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(_FakeThreadAdapter.instances) == 1
    thread_adapter = _FakeThreadAdapter.instances[0]
    assert thread_adapter.writes == []


@pytest.mark.asyncio
async def test_worker_none_threading_does_not_open_thread_adapter() -> None:
    _FakeThreadAdapter.instances.clear()
    adapter = _FakeLLMAdapter()
    worker = Worker(
        queue="tasks",
        llm_adapter=adapter,
        threading_mode="none",
        thread_dir="/threads",
    )
    room = _FakeRoom()
    worker._room = room
    worker._threading_helper._thread_adapter_type = _FakeThreadAdapter

    result = await worker.process_message(
        chat_context=AgentSessionContext(system_role=None),
        message={"prompt": "No persistence"},
        toolkits=[],
    )

    assert result == "assistant response"
    assert len(adapter.calls) == 1
    assert _FakeThreadAdapter.instances == []
    assert room.sync.open_calls == []
    assert room.sync.close_calls == []



import uuid
from unittest import mock

import pytest
from pydantic import BaseModel

from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.legacy_chat_channel import LegacyChatChannel as ChatChannel
from meshagent.agents.thread_schema import thread_list_schema
from meshagent.agents.messages import (
    AGENT_EVENT_FILE_CONTENT_DELTA,
    AGENT_EVENT_FILE_CONTENT_ENDED,
    AGENT_EVENT_FILE_CONTENT_STARTED,
    AGENT_EVENT_THREAD_CLEARED,
    AGENT_EVENT_TEXT_CONTENT_DELTA,
    AGENT_EVENT_TEXT_CONTENT_ENDED,
    AGENT_EVENT_TEXT_CONTENT_STARTED,
    AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
    AGENT_EVENT_TURN_ENDED,
    AGENT_EVENT_TURN_STARTED,
    AGENT_MESSAGE_THREAD_CLEAR,
    AGENT_MESSAGE_TOOL_CALL_APPROVE,
    AGENT_MESSAGE_TURN_INTERRUPT,
    AGENT_MESSAGE_TURN_START,
    AGENT_MESSAGE_TURN_STEER,
    AgentFileContent,
    AgentFileContentDelta,
    AgentFileContentEnded,
    AgentFileContentStarted,
    AgentTextContent,
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    AgentToolCallApprovalRequested,
    ApproveAgentToolCall,
    ClearThread,
    ThreadCleared,
    TurnEnded,
    TurnInterrupt,
    TurnStart,
    TurnStarted,
    TurnSteer,
)
from meshagent.agents.process import AgentSupervisor, Message
from meshagent.api import Participant, RoomException, RoomMessage
from meshagent.api.messaging import EmptyContent, JsonContent
from meshagent.tools import ToolContext, ToolkitBuilder


class _FakeParticipant(Participant):
    def __init__(self, *, name: str, participant_id: str) -> None:
        super().__init__(id=participant_id, attributes={"name": name})


class _FakeLocalParticipant(_FakeParticipant):
    def __init__(self) -> None:
        super().__init__(name="assistant", participant_id="assistant-id")
        self.set_attribute_calls: list[tuple[str, object]] = []

    async def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value
        self.set_attribute_calls.append((name, value))


class _FakeThreadListElement:
    def __init__(self, *, tag_name: str, attributes: dict[str, str]) -> None:
        self.tag_name = tag_name
        self._attributes = dict(attributes)

    def get_attribute(self, name: str):
        return self._attributes.get(name)

    def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value


class _FakeThreadListRoot:
    def __init__(self) -> None:
        self._children: list[_FakeThreadListElement] = []

    def get_children(self) -> list[_FakeThreadListElement]:
        return [*self._children]

    def append_child(
        self,
        *,
        tag_name: str,
        attributes: dict[str, str],
    ) -> _FakeThreadListElement:
        element = _FakeThreadListElement(tag_name=tag_name, attributes=attributes)
        self._children.append(element)
        return element


class _FakeThreadListDocument:
    def __init__(self) -> None:
        self.root = _FakeThreadListRoot()


class _FakeSync:
    def __init__(self) -> None:
        self.document = _FakeThreadListDocument()
        self.open_calls: list[dict[str, object]] = []
        self.close_calls: list[str] = []

    async def open(
        self,
        *,
        path: str,
        schema=None,
    ) -> _FakeThreadListDocument:
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str) -> None:
        self.close_calls.append(path)


class _FakeStorage:
    def __init__(self, *, existing_paths: set[str] | None = None) -> None:
        self._existing_paths = set(existing_paths or [])
        self.exists_calls: list[str] = []

    async def exists(self, *, path: str) -> bool:
        self.exists_calls.append(path)
        return path in self._existing_paths


class _FakeMessaging:
    def __init__(
        self,
        *,
        participants: list[Participant] | None = None,
        is_enabled: bool = False,
    ) -> None:
        self._participants = {
            participant.id: participant for participant in participants or []
        }
        self._handlers: dict[str, list] = {}
        self._is_enabled = is_enabled
        self.enable_calls = 0
        self.sent_messages: list[dict] = []

    @property
    def is_enabled(self) -> bool:
        return self._is_enabled

    async def enable(self) -> None:
        self.enable_calls += 1
        self._is_enabled = True

    def on(self, event_name: str, func) -> None:
        handlers = self._handlers.setdefault(event_name, [])
        handlers.append(func)

    def off(self, event_name: str, func) -> None:
        handlers = self._handlers.get(event_name)
        if handlers is None:
            return
        handlers.remove(func)

    def get_participant(self, participant_id: str) -> Participant | None:
        return self._participants.get(participant_id)

    def send_message_nowait(
        self,
        *,
        to: Participant,
        type: str,
        message: dict,
        attachment=None,
    ) -> None:
        del attachment
        self.sent_messages.append({"to": to, "type": type, "message": message})

    def emit_message(self, message: RoomMessage) -> None:
        for handler in self._handlers.get("message", []):
            handler(message=message)

    def remove_participant(self, participant_id: str) -> None:
        self._participants.pop(participant_id, None)


class _FakeRoom:
    def __init__(
        self,
        *,
        participants: list[Participant] | None = None,
        messaging_enabled: bool = False,
        sync: _FakeSync | None = None,
        storage: _FakeStorage | None = None,
    ) -> None:
        self.local_participant = _FakeLocalParticipant()
        self.messaging = _FakeMessaging(
            participants=participants,
            is_enabled=messaging_enabled,
        )
        self.sync = sync if sync is not None else _FakeSync()
        self.storage = storage if storage is not None else _FakeStorage()


class _FakeToolkitConfig(BaseModel):
    name: str


class _FakeToolkitBuilder(ToolkitBuilder):
    def __init__(self, *, name: str) -> None:
        super().__init__(name=name, type=_FakeToolkitConfig)

    async def make(self, *, room, model: str, config: _FakeToolkitConfig):
        del room
        del model
        del config
        raise AssertionError(
            "toolkit builder should not be called in chat channel test"
        )


class _RecordingSupervisor(AgentSupervisor):
    def __init__(self) -> None:
        super().__init__()
        self.sent: list[Message] = []

    def send(self, message: Message) -> None:
        self.sent.append(message)


class _FakeThreadNameAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str) -> None:
        self.generated_thread_name = generated_thread_name
        self.prompts: list[str] = []

    def default_model(self) -> str:
        return "thread-name-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options=None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        self.prompts = [
            message["content"]
            for message in context.messages
            if isinstance(message, dict) and isinstance(message.get("content"), str)
        ]
        return {"thread_name": self.generated_thread_name}


def _assert_uuid_thread_path(*, path: str, prefix: str) -> None:
    assert path.startswith(prefix)
    assert path.endswith(".thread")
    basename = path[len(prefix) : -len(".thread")]
    parsed = uuid.UUID(basename)
    assert str(parsed) == basename


@pytest.mark.asyncio
async def test_chat_channel_exposes_chat_toolkits_and_new_thread_emits_turn_start() -> (
    None
):
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    sync = _FakeSync()
    room = _FakeRoom(
        participants=[caller],
        messaging_enabled=True,
        sync=sync,
    )
    channel = ChatChannel(
        room=room,
        thread_dir="/threads/chat",
        toolkit_builders=[_FakeToolkitBuilder(name="search")],
    )
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        agent_toolkits = channel.get_agent_toolkits()
        assert len(agent_toolkits) == 1
        agent_tool_names = {tool.name for tool in agent_toolkits[0].tools}
        assert agent_tool_names == {
            "new_thread",
            "attach_file",
            "list_threads",
            "grep_thread_list",
        }

        remote_toolkit = channel.make_remote_toolkit()
        assert {tool.name for tool in remote_toolkit.tools} == agent_tool_names
        assert remote_toolkit.validation_mode == "content_types"

        new_thread_tool = next(
            tool for tool in agent_toolkits[0].tools if tool.name == "new_thread"
        )
        tools_schema = new_thread_tool.input_schema["properties"]["message"][
            "properties"
        ]["tools"]
        assert len(tools_schema["anyOf"]) == 2
        assert tools_schema["anyOf"][0]["type"] == "array"
        assert tools_schema["anyOf"][0]["items"]["type"] == "object"
        assert (
            tools_schema["anyOf"][0]["items"]["properties"]["name"]["type"] == "string"
        )
        assert tools_schema["anyOf"][1] == {"type": "null"}
        context = ToolContext(room=room, caller=caller)
        result = await new_thread_tool.execute(
            context=context,
            message={
                "text": "Plan this friendly thread",
                "attachments": [{"path": "uploads/plan.md"}],
                "tools": [{"name": "search"}],
            },
        )

        assert isinstance(result, JsonContent)
        result_path = result.json["path"]
        assert result.json["name"] == "Plan this friendly thread"
        assert isinstance(result_path, str)
        _assert_uuid_thread_path(path=result_path, prefix="/threads/chat/")

        assert len(supervisor.sent) == 1
        sent = supervisor.sent[0]
        assert sent.sender is caller
        assert sent.source is channel
        turn = sent.data
        assert isinstance(turn, TurnStart)
        assert turn.type == AGENT_MESSAGE_TURN_START
        assert turn.thread_id == result_path
        assert turn.toolkits == [{"name": "search"}]
        assert turn.content == [
            AgentTextContent(type="text", text="Plan this friendly thread"),
            AgentFileContent(type="file", url="room:///uploads/plan.md"),
        ]

        entries = sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("path") == result_path
        assert entries[0].get_attribute("name") == "Plan this friendly thread"

        list_threads_tool = next(
            tool for tool in agent_toolkits[0].tools if tool.name == "list_threads"
        )
        list_result = await list_threads_tool.execute(
            context=context,
            limit=20,
            offset=0,
        )
        assert isinstance(list_result, JsonContent)
        assert list_result.json["total"] == 1
        assert list_result.json["threads"][0]["path"] == result_path
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_new_thread_uses_message_text_and_attachment_names_for_llm_naming() -> (
    None
):
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    sync = _FakeSync()
    adapter = _FakeThreadNameAdapter(generated_thread_name="Release Plan")
    room = _FakeRoom(
        participants=[caller],
        messaging_enabled=True,
        sync=sync,
    )
    channel = ChatChannel(
        room=room,
        thread_dir="/threads/chat",
        llm_adapter=adapter,
    )
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        new_thread_tool = next(
            tool
            for tool in channel.get_agent_toolkits()[0].tools
            if tool.name == "new_thread"
        )
        result = await new_thread_tool.execute(
            context=ToolContext(room=room, caller=caller),
            message={
                "text": "Plan the release work",
                "attachments": [
                    {"path": "uploads/release-plan.md"},
                    {"path": "uploads/screenshot.png"},
                ],
            },
        )

        assert isinstance(result, JsonContent)
        assert result.json["name"] == "Release Plan"
        assert adapter.prompts == [
            "Message:\nPlan the release work\n\nAttachments:\n- release-plan.md\n- screenshot.png"
        ]

        entries = sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Release Plan"
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_attach_file_emits_file_content_events() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(
        participants=[caller],
        messaging_enabled=True,
        storage=_FakeStorage(existing_paths={"docs/report.pdf"}),
    )
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        attach_file_tool = next(
            tool
            for tool in channel.get_agent_toolkits()[0].tools
            if tool.name == "attach_file"
        )

        result = await attach_file_tool.execute(
            context=ToolContext(
                room=room,
                caller=caller,
                caller_context={
                    "thread_id": "/threads/test.thread",
                    "turn_id": "turn-1",
                },
            ),
            path="docs/report.pdf",
        )

        assert isinstance(result, EmptyContent)
        assert room.storage.exists_calls == ["docs/report.pdf"]
        assert len(supervisor.sent) == 3

        started = supervisor.sent[0]
        delta = supervisor.sent[1]
        ended = supervisor.sent[2]
        assert started.sender is caller
        assert delta.sender is caller
        assert ended.sender is caller

        started_payload = started.data
        delta_payload = delta.data
        ended_payload = ended.data
        assert isinstance(started_payload, AgentFileContentStarted)
        assert started_payload.type == AGENT_EVENT_FILE_CONTENT_STARTED
        assert started_payload.thread_id == "/threads/test.thread"
        assert started_payload.turn_id == "turn-1"
        assert isinstance(delta_payload, AgentFileContentDelta)
        assert delta_payload.type == AGENT_EVENT_FILE_CONTENT_DELTA
        assert delta_payload.thread_id == "/threads/test.thread"
        assert delta_payload.turn_id == "turn-1"
        assert delta_payload.url == "room:///docs/report.pdf"
        assert isinstance(ended_payload, AgentFileContentEnded)
        assert ended_payload.type == AGENT_EVENT_FILE_CONTENT_ENDED
        assert ended_payload.thread_id == "/threads/test.thread"
        assert ended_payload.turn_id == "turn-1"
        assert started_payload.item_id == delta_payload.item_id
        assert delta_payload.item_id == ended_payload.item_id
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_attach_file_raises_for_missing_room_file() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        attach_file_tool = next(
            tool
            for tool in channel.get_agent_toolkits()[0].tools
            if tool.name == "attach_file"
        )

        with pytest.raises(
            RoomException,
            match=r"attach_file could not find a room file at docs/missing\.pdf",
        ):
            await attach_file_tool.execute(
                context=ToolContext(
                    room=room,
                    caller=caller,
                    caller_context={
                        "thread_id": "/threads/test.thread",
                        "turn_id": "turn-1",
                    },
                ),
                path="docs/missing.pdf",
            )

        assert room.storage.exists_calls == ["docs/missing.pdf"]
        assert supervisor.sent == []
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_default_new_exposes_thread_list_tools_without_explicit_thread_dir() -> (
    None
):
    sync = _FakeSync()
    room = _FakeRoom(messaging_enabled=True, sync=sync)
    channel = ChatChannel(room=room, threading_mode="default-new")
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        tool_names = {tool.name for tool in channel.get_agent_toolkits()[0].tools}
        assert tool_names == {
            "new_thread",
            "attach_file",
            "list_threads",
            "grep_thread_list",
        }
        assert channel.make_remote_toolkit().validation_mode == "content_types"
        assert (
            "meshagent.chatbot.thread-list",
            ".threads/assistant/index.threadl",
        ) in room.local_participant.set_attribute_calls
        assert (
            "meshagent.chatbot.thread-dir",
            ".threads/assistant",
        ) in room.local_participant.set_attribute_calls
        assert sync.open_calls == [
            {
                "path": ".threads/assistant/index.threadl",
                "schema": thread_list_schema,
            }
        ]
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_enables_messaging_and_translates_chat_messages() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller])
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        assert room.messaging.enable_calls == 1

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="chat",
                message={
                    "path": "/threads/test.thread",
                    "text": "hello",
                    "attachments": [{"path": "docs/report.pdf"}],
                    "tools": [{"name": "search"}],
                    "model": "gpt-5",
                },
            )
        )

        assert len(supervisor.sent) == 1
        sent = supervisor.sent[0]
        assert sent.sender is caller
        assert sent.source is channel

        turn = sent.data
        assert isinstance(turn, TurnStart)
        assert turn.type == AGENT_MESSAGE_TURN_START
        assert turn.thread_id == "/threads/test.thread"
        assert turn.model == "gpt-5"
        assert turn.toolkits == [{"name": "search"}]
        assert turn.content == [
            AgentTextContent(type="text", text="hello"),
            AgentFileContent(type="file", url="room:///docs/report.pdf"),
        ]
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_preserves_existing_attachment_urls() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="chat",
                message={
                    "path": "/threads/test.thread",
                    "attachments": [
                        {"path": "room://docs/report.pdf"},
                        {"path": "https://example.com/image.png"},
                    ],
                },
            )
        )

        assert len(supervisor.sent) == 1
        turn = supervisor.sent[0].data
        assert isinstance(turn, TurnStart)
        assert turn.content == [
            AgentFileContent(type="file", url="room://docs/report.pdf"),
            AgentFileContent(type="file", url="https://example.com/image.png"),
        ]
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_tracks_turn_state_for_steer_cancel_and_approval() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        assert room.messaging.enable_calls == 0

        await channel.on_message(
            Message(
                data=TurnStarted(
                    type=AGENT_EVENT_TURN_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    source_message_id="source-1",
                )
            )
        )

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="steer",
                message={
                    "path": "/threads/test.thread",
                    "text": "keep going",
                },
            )
        )

        steer_message = supervisor.sent[0].data
        assert isinstance(steer_message, TurnSteer)
        assert steer_message.type == AGENT_MESSAGE_TURN_STEER
        assert steer_message.thread_id == "/threads/test.thread"
        assert steer_message.turn_id == "turn-1"
        assert steer_message.content == [
            AgentTextContent(type="text", text="keep going")
        ]

        await channel.on_message(
            Message(
                data=AgentToolCallApprovalRequested(
                    type=AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="approval-1",
                    toolkit="filesystem",
                    tool="delete",
                    arguments={"path": "/tmp/file"},
                )
            )
        )

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="approved",
                message={
                    "path": "/threads/test.thread",
                    "approval_id": "approval-1",
                },
            )
        )

        approval_message = supervisor.sent[1].data
        assert isinstance(approval_message, ApproveAgentToolCall)
        assert approval_message.type == AGENT_MESSAGE_TOOL_CALL_APPROVE
        assert approval_message.thread_id == "/threads/test.thread"
        assert approval_message.turn_id == "turn-1"
        assert approval_message.item_id == "approval-1"

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="cancel",
                message={"path": "/threads/test.thread"},
            )
        )

        interrupt_message = supervisor.sent[2].data
        assert isinstance(interrupt_message, TurnInterrupt)
        assert interrupt_message.type == AGENT_MESSAGE_TURN_INTERRUPT
        assert interrupt_message.thread_id == "/threads/test.thread"
        assert interrupt_message.turn_id == "turn-1"

        await channel.on_message(
            Message(
                data=TurnEnded(
                    type=AGENT_EVENT_TURN_ENDED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    error=None,
                )
            )
        )

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="steer",
                message={
                    "path": "/threads/test.thread",
                    "text": "one more thing",
                },
            )
        )

        assert len(supervisor.sent) == 3
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_sends_completed_text_to_open_participants() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="opened",
                message={"path": "/threads/test.thread"},
            )
        )

        assert supervisor.sent == []

        await channel.on_message(
            Message(
                data=AgentTextContentStarted(
                    type=AGENT_EVENT_TEXT_CONTENT_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-1",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentDelta(
                    type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-1",
                    text="hello",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentEnded(
                    type=AGENT_EVENT_TEXT_CONTENT_ENDED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-1",
                )
            )
        )

        assert room.messaging.sent_messages == [
            {
                "to": caller,
                "type": "chat",
                "message": {
                    "path": "/threads/test.thread",
                    "text": "hello",
                },
            }
        ]

        room.messaging.remove_participant(caller.id)

        await channel.on_message(
            Message(
                data=AgentTextContentStarted(
                    type=AGENT_EVENT_TEXT_CONTENT_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-2",
                    item_id="assistant-2",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentDelta(
                    type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                    thread_id="/threads/test.thread",
                    turn_id="turn-2",
                    item_id="assistant-2",
                    text="goodbye",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentEnded(
                    type=AGENT_EVENT_TEXT_CONTENT_ENDED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-2",
                    item_id="assistant-2",
                )
            )
        )

        assert len(room.messaging.sent_messages) == 1
        assert channel._open_participant_ids_by_thread == {}
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_sends_completed_file_to_open_participants() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="opened",
                message={"path": "/threads/test.thread"},
            )
        )

        await channel.on_message(
            Message(
                data=AgentFileContentStarted(
                    type=AGENT_EVENT_FILE_CONTENT_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-file-1",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentFileContentDelta(
                    type=AGENT_EVENT_FILE_CONTENT_DELTA,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-file-1",
                    url="room:///docs/report.pdf",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentFileContentEnded(
                    type=AGENT_EVENT_FILE_CONTENT_ENDED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-file-1",
                )
            )
        )

        assert room.messaging.sent_messages == [
            {
                "to": caller,
                "type": "chat",
                "message": {
                    "path": "/threads/test.thread",
                    "attachments": [{"path": "room:///docs/report.pdf"}],
                },
            }
        ]
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_translates_clear_messages() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="clear",
                message={"path": "/threads/test.thread"},
            )
        )

        assert len(supervisor.sent) == 1
        clear_message = supervisor.sent[0].data
        assert isinstance(clear_message, ClearThread)
        assert clear_message.type == AGENT_MESSAGE_THREAD_CLEAR
        assert clear_message.thread_id == "/threads/test.thread"
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_notifies_open_participants_when_thread_is_cleared() -> None:
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    room = _FakeRoom(participants=[caller], messaging_enabled=True)
    channel = ChatChannel(room=room)
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="opened",
                message={"path": "/threads/test.thread"},
            )
        )

        await channel.on_message(
            Message(
                data=TurnStarted(
                    type=AGENT_EVENT_TURN_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    source_message_id="source-1",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentToolCallApprovalRequested(
                    type=AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="approval-1",
                    toolkit="filesystem",
                    tool="delete",
                    arguments={"path": "/tmp/file"},
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentStarted(
                    type=AGENT_EVENT_TEXT_CONTENT_STARTED,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-1",
                )
            )
        )
        await channel.on_message(
            Message(
                data=AgentTextContentDelta(
                    type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                    thread_id="/threads/test.thread",
                    turn_id="turn-1",
                    item_id="assistant-1",
                    text="partial",
                )
            )
        )

        await channel.on_message(
            Message(
                data=ThreadCleared(
                    type=AGENT_EVENT_THREAD_CLEARED,
                    thread_id="/threads/test.thread",
                    source_message_id="clear-source-1",
                )
            )
        )

        assert room.messaging.sent_messages == [
            {
                "to": caller,
                "type": "cleared",
                "message": {
                    "path": "/threads/test.thread",
                },
            }
        ]
        assert channel._active_turn_ids_by_thread == {}
        assert channel._pending_approval_turn_ids_by_thread == {}
        assert channel._active_text_by_thread == {}
        assert channel._open_participant_ids_by_thread == {
            "/threads/test.thread": {caller.id}
        }
    finally:
        await channel.stop(supervisor)


@pytest.mark.asyncio
async def test_chat_channel_sets_threading_attributes_tracks_thread_list_and_reports_tool_providers() -> (
    None
):
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    sync = _FakeSync()
    room = _FakeRoom(
        participants=[caller],
        messaging_enabled=True,
        sync=sync,
    )
    channel = ChatChannel(
        room=room,
        threading_mode="default-new",
        thread_dir="/threads/chat",
        toolkit_builders=[
            _FakeToolkitBuilder(name="search"),
            _FakeToolkitBuilder(name="shell"),
        ],
    )
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        assert room.local_participant.set_attribute_calls == [
            ("meshagent.chatbot.threading", "default-new"),
            ("meshagent.chatbot.thread-dir", "/threads/chat"),
            ("meshagent.chatbot.thread-list", "/threads/chat/index.threadl"),
            ("empty_state_title", "How can I help you?"),
        ]
        assert sync.open_calls == [
            {
                "path": "/threads/chat/index.threadl",
                "schema": thread_list_schema,
            }
        ]

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="opened",
                message={"path": "/threads/chat/example.thread"},
            )
        )

        assert sync.document.root.get_children() == []

        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="get_thread_toolkit_builders",
                message={"path": "/threads/chat/example.thread"},
            )
        )

        assert room.messaging.sent_messages == [
            {
                "to": caller,
                "type": "set_thread_tool_providers",
                "message": {
                    "path": "/threads/chat/example.thread",
                    "tool_providers": [
                        {"name": "search"},
                        {"name": "shell"},
                    ],
                },
            }
        ]
    finally:
        await channel.stop(supervisor)

    assert sync.close_calls == ["/threads/chat/index.threadl"]


@pytest.mark.asyncio
async def test_chat_channel_does_not_bump_thread_index_on_open_or_tool_provider_request() -> (
    None
):
    caller = _FakeParticipant(name="caller", participant_id="caller-id")
    sync = _FakeSync()
    existing_entry = sync.document.root.append_child(
        tag_name="thread",
        attributes={
            "path": "/threads/chat/example.thread",
            "name": "Example",
            "created_at": "2024-01-01T00:00:00Z",
            "modified_at": "2024-01-01T00:00:00Z",
        },
    )
    room = _FakeRoom(
        participants=[caller],
        messaging_enabled=True,
        sync=sync,
    )
    channel = ChatChannel(
        room=room,
        threading_mode="default-new",
        thread_dir="/threads/chat",
        toolkit_builders=[_FakeToolkitBuilder(name="search")],
    )
    supervisor = _RecordingSupervisor()

    await channel.start(supervisor)
    try:
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="opened",
                message={"path": "/threads/chat/example.thread"},
            )
        )
        room.messaging.emit_message(
            RoomMessage(
                from_participant_id=caller.id,
                type="get_thread_toolkit_builders",
                message={"path": "/threads/chat/example.thread"},
            )
        )

        assert existing_entry.get_attribute("modified_at") == "2024-01-01T00:00:00Z"
    finally:
        await channel.stop(supervisor)


def test_chat_channel_bump_thread_forces_modified_at_forward_when_clock_does_not_advance() -> (
    None
):
    existing_modified_at = "2024-01-01T00:00:00Z"
    sync = _FakeSync()
    existing_entry = sync.document.root.append_child(
        tag_name="thread",
        attributes={
            "path": "/threads/chat/example.thread",
            "name": "Example",
            "created_at": existing_modified_at,
            "modified_at": existing_modified_at,
        },
    )
    room = _FakeRoom(sync=sync)
    channel = ChatChannel(
        room=room,
        threading_mode="default-new",
        thread_dir="/threads/chat",
    )
    channel._thread_list_document = sync.document

    with mock.patch.object(channel, "_utc_now_iso", return_value=existing_modified_at):
        channel.bump_thread(path="/threads/chat/example.thread")

    updated_modified_at = existing_entry.get_attribute("modified_at")
    assert isinstance(updated_modified_at, str)
    assert channel._parse_iso_datetime(
        value=updated_modified_at
    ) > channel._parse_iso_datetime(value=existing_modified_at)



import pytest
import base64
from typing import Optional

import meshagent.agents.thread_adapter as thread_adapter_module
from meshagent.agents.thread_adapter import ThreadAdapter


class _FakeThreadAdapter(ThreadAdapter):
    def __init__(self) -> None:
        super().__init__(room=object(), path="/threads/test")  # type: ignore[arg-type]
        self.started = 0
        self.stopped = 0

    async def start(self) -> None:
        self.started += 1

    async def stop(self) -> None:
        self.stopped += 1

    async def handle_custom_event(self, *, messages, event) -> None:
        del messages
        del event

    async def _process_llm_events(self) -> None:
        return None


@pytest.mark.asyncio
async def test_thread_adapter_async_manager_calls_start_and_stop() -> None:
    adapter = _FakeThreadAdapter()

    async with adapter:
        assert adapter.started == 1
        assert adapter.stopped == 0

    assert adapter.started == 1
    assert adapter.stopped == 1


class _FakeSync:
    def __init__(self) -> None:
        self.sync_calls: list[dict] = []
        self.close_calls: list[str] = []

    async def sync(self, *, path: str, data: bytes) -> None:
        self.sync_calls.append({"path": path, "data": data})

    async def close(self, *, path: str) -> None:
        self.close_calls.append(path)


class _FakeRoom:
    def __init__(self) -> None:
        self.sync = _FakeSync()


class _FakeMeshDocument:
    def __init__(self, *, state: Optional[bytes] = None) -> None:
        self._state = state if state is not None else b""

    def get_state(self, vector: bytes | None = None) -> bytes:
        del vector
        return self._state


class _BaseStopThreadAdapter(ThreadAdapter):
    async def handle_custom_event(self, *, messages, event) -> None:
        del messages
        del event

    async def _process_llm_events(self) -> None:
        return None


@pytest.mark.asyncio
async def test_thread_adapter_stop_flushes_state_before_close(monkeypatch) -> None:
    async def _fast_sleep(delay: float) -> None:
        del delay

    monkeypatch.setattr(thread_adapter_module.asyncio, "sleep", _fast_sleep)

    room = _FakeRoom()
    adapter = _BaseStopThreadAdapter(room=room, path="/threads/test")
    adapter._thread = _FakeMeshDocument(state=b"state")

    await adapter.stop()

    assert room.sync.sync_calls == [
        {"path": "/threads/test", "data": base64.standard_b64encode(b"state")}
    ]
    assert room.sync.close_calls == ["/threads/test"]



import asyncio
from datetime import datetime, timezone
import uuid

import pytest

from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.messages import TurnStart
from meshagent.agents.process import Message
from meshagent.agents.queue_channel import QueueChannel
from meshagent.agents.thread_schema import thread_list_schema
from meshagent.api import Participant
from meshagent.api.messaging import FileContent


class _FakeLocalParticipant(Participant):
    def __init__(self) -> None:
        super().__init__(id="assistant-id", attributes={"name": "assistant"})
        self.set_attribute_calls: list[tuple[str, object]] = []

    async def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value
        self.set_attribute_calls.append((name, value))


class _FakeQueues:
    def __init__(self) -> None:
        self._queue: asyncio.Queue[object] = asyncio.Queue()
        self.receive_calls: list[dict[str, object]] = []

    async def receive(self, *, name: str, create: bool, wait: bool):
        self.receive_calls.append({"name": name, "create": create, "wait": wait})
        return await self._queue.get()

    async def push(self, payload: object) -> None:
        await self._queue.put(payload)


class _FakeThreadListElement:
    def __init__(self, *, tag_name: str, attributes: dict[str, str]) -> None:
        self.tag_name = tag_name
        self._attributes = dict(attributes)

    def get_attribute(self, name: str):
        return self._attributes.get(name)

    def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value


class _FakeThreadListRoot:
    def __init__(self) -> None:
        self._children: list[_FakeThreadListElement] = []

    def get_children(self) -> list[_FakeThreadListElement]:
        return [*self._children]

    def append_child(
        self,
        *,
        tag_name: str,
        attributes: dict[str, str],
    ) -> _FakeThreadListElement:
        element = _FakeThreadListElement(tag_name=tag_name, attributes=attributes)
        self._children.append(element)
        return element


class _FakeThreadListDocument:
    def __init__(self) -> None:
        self.root = _FakeThreadListRoot()


class _FakeSync:
    def __init__(self) -> None:
        self.document = _FakeThreadListDocument()
        self.open_calls: list[dict[str, object]] = []
        self.close_calls: list[str] = []

    async def open(self, *, path: str, schema=None) -> _FakeThreadListDocument:
        self.open_calls.append({"path": path, "schema": schema})
        return self.document

    async def close(self, *, path: str) -> None:
        self.close_calls.append(path)


class _FakeStorage:
    def __init__(self, *, files: dict[str, bytes] | None = None) -> None:
        self._files = dict(files or {})
        self.exists_calls: list[str] = []
        self.download_calls: list[str] = []

    async def exists(self, *, path: str) -> bool:
        self.exists_calls.append(path)
        return path in self._files

    async def download(self, *, path: str) -> FileContent:
        self.download_calls.append(path)
        return FileContent(
            data=self._files[path],
            name=path.rsplit("/", 1)[-1],
            mime_type="text/plain",
        )


class _FakeRoom:
    def __init__(self, *, files: dict[str, bytes] | None = None) -> None:
        self.local_participant = _FakeLocalParticipant()
        self.queues = _FakeQueues()
        self.sync = _FakeSync()
        self.storage = _FakeStorage(files=files)


class _RecordingSupervisor:
    def __init__(self) -> None:
        self.sent: list[Message] = []

    def send(self, message: Message) -> None:
        self.sent.append(message)


class _FakeThreadNameAdapter(LLMAdapter):
    def __init__(self, *, generated_thread_name: str) -> None:
        self.generated_thread_name = generated_thread_name
        self.prompts: list[str] = []

    def default_model(self) -> str:
        return "thread-name-model"

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options=None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        self.prompts = [
            message["content"]
            for message in context.messages
            if isinstance(message, dict) and isinstance(message.get("content"), str)
        ]
        return {"thread_name": self.generated_thread_name}


async def _drain() -> None:
    await asyncio.sleep(0)
    await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_queue_channel_emits_turn_start_from_prompt_and_path() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push(
            {
                "prompt": "Process webhook payload",
                "path": ".threads/jobs/webhook.thread",
                "model": "gpt-5.4",
                "instructions": "Be concise",
                "tools": [{"name": "search"}],
                "sender_name": "Webhook",
            }
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.thread_id == ".threads/jobs/webhook.thread"
        assert outbound.data.model == "gpt-5.4"
        assert outbound.data.instructions == "Be concise"
        assert outbound.data.toolkits == [{"name": "search"}]
        assert outbound.data.content[0].text == "Process webhook payload"
        assert outbound.sender is not None
        assert outbound.sender.get_attribute("name") == "Webhook"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_generates_thread_path_and_uses_payload_json_when_prompt_missing() -> (
    None
):
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push({"body": "hello", "value": 3})
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.thread_id.startswith(".threads/assistant/")
        assert outbound.data.thread_id.endswith(".thread")
        assert '"body": "hello"' in outbound.data.content[0].text
        assert outbound.sender is None
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_supports_string_messages() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs", thread_dir=".threads/queue")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push("Do the thing")
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.content[0].text == "Do the thing"
        assert outbound.data.thread_id.startswith(".threads/queue/")
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_supports_structured_prompt_with_room_file_include() -> (
    None
):
    room = _FakeRoom(files={"prompts/heartbeat.md": b"Heartbeat instructions"})
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push(
            {
                "thread_id": "/threads/heartbeats/current.thread",
                "prompt": [
                    {"type": "file", "url": "room:///prompts/heartbeat.md"},
                    {"type": "text", "text": "Summarize the room status"},
                    {"type": "file", "url": "https://example.com/docs/report.md"},
                ],
            }
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.thread_id == "/threads/heartbeats/current.thread"
        assert [item.type for item in outbound.data.content] == ["text", "text", "file"]
        assert outbound.data.content[0].text == "Heartbeat instructions"
        assert outbound.data.content[1].text == "Summarize the room status"
        assert outbound.data.content[2].url == "https://example.com/docs/report.md"
        assert room.storage.exists_calls == ["prompts/heartbeat.md"]
        assert room.storage.download_calls == ["prompts/heartbeat.md"]
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_preserves_structured_content_room_files() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push(
            {
                "thread_id": "/threads/uploads/current.thread",
                "content": [
                    {"type": "file", "url": "room:///docs/report.md"},
                ],
            }
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.content[0].url == "room:///docs/report.md"
        assert room.storage.exists_calls == []
        assert room.storage.download_calls == []
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_expands_datetime_tokens_in_thread_id(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    fixed_now = datetime(2026, 1, 2, 3, 4, 5, tzinfo=timezone.utc)
    monkeypatch.setattr(QueueChannel, "_now", lambda self: fixed_now)

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push(
            {
                "prompt": "Run the hourly heartbeat",
                "thread_id": "/threads/{YYYY}/{MM}/{DD}/{HH}/{mm}/heartbeat.thread",
            }
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert outbound.data.thread_id == "/threads/2026/01/02/03/04/heartbeat.thread"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_expands_case_insensitive_datetime_aliases_in_thread_id(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(room=room, queue_name="jobs")
    fixed_now = datetime(2026, 1, 2, 3, 4, 5, tzinfo=timezone.utc)
    monkeypatch.setattr(QueueChannel, "_now", lambda self: fixed_now)

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push(
            {
                "prompt": "Run the hourly heartbeat",
                "thread_id": "/threads/{yEaR}/{mOnTh}/{DaY}/{hOuR}/{mInUtE}/{sEcOnD}/heartbeat.thread",
            }
        )
        await _drain()

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert (
            outbound.data.thread_id == "/threads/2026/01/02/03/04/05/heartbeat.thread"
        )
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_queue_channel_default_new_indexes_new_threads_from_prompt() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    channel = QueueChannel(
        room=room,
        queue_name="jobs",
        threading_mode="default-new",
    )
    expected_uuid = uuid.UUID("12345678-1234-5678-1234-567812345678")

    original_uuid4 = uuid.uuid4
    uuid.uuid4 = lambda: expected_uuid
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push("follow up on billing issue")
        await _drain()

        assert room.local_participant.set_attribute_calls == [
            ("meshagent.chatbot.threading", "default-new"),
            ("meshagent.chatbot.thread-dir", ".threads/assistant"),
            ("meshagent.chatbot.thread-list", ".threads/assistant/index.threadl"),
        ]
        assert room.sync.open_calls == [
            {
                "path": ".threads/assistant/index.threadl",
                "schema": thread_list_schema,
            }
        ]
        assert room.storage.exists_calls == [
            ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        ]

        assert len(supervisor.sent) == 1
        outbound = supervisor.sent[0]
        assert isinstance(outbound.data, TurnStart)
        assert (
            outbound.data.thread_id
            == ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        )

        entries = room.sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Follow Up On Billing Issue"
        assert (
            entries[0].get_attribute("path")
            == ".threads/assistant/12345678-1234-5678-1234-567812345678.thread"
        )
    finally:
        uuid.uuid4 = original_uuid4
        await channel.stop(supervisor)  # type: ignore[arg-type]

    assert room.sync.close_calls == [".threads/assistant/index.threadl"]


@pytest.mark.asyncio
async def test_queue_channel_default_new_uses_llm_thread_name() -> None:
    room = _FakeRoom()
    supervisor = _RecordingSupervisor()
    adapter = _FakeThreadNameAdapter(generated_thread_name="Billing Follow Up")
    channel = QueueChannel(
        room=room,
        queue_name="jobs",
        threading_mode="default-new",
        llm_adapter=adapter,
    )

    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        await room.queues.push("follow up on billing issue")
        await _drain()

        assert adapter.prompts == ["Message:\nfollow up on billing issue"]
        entries = room.sync.document.root.get_children()
        assert len(entries) == 1
        assert entries[0].get_attribute("name") == "Billing Follow Up"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]



from __future__ import annotations

import logging
import posixpath
import re
import uuid
from datetime import datetime, timedelta, timezone
from pathlib import PurePosixPath
from typing import Any, Sequence
from urllib.parse import urlparse

from meshagent.api import Element, MeshDocument, Participant, RoomClient

from .adapter import LLMAdapter
from .process import Channel
from .thread_schema import thread_list_schema

logger = logging.getLogger("threaded-channel")

DEFAULT_CHANNEL_THREAD_NAME_RULES = [
    "generate a concise, friendly title for this chat thread",
    "return only a thread_name value suitable for display in a thread list",
    "thread_name should be 2-6 words and topic-focused",
    "use normal capitalization and spaces, and do not include a .thread extension",
]


class ThreadedChannel(Channel):
    def __init__(
        self,
        *,
        room: RoomClient,
        threading_mode: str | None = None,
        thread_dir: str | None = None,
        llm_adapter: LLMAdapter | None = None,
        thread_name_rules: Sequence[str] | None = None,
    ) -> None:
        super().__init__()
        self._room = room
        self._threading_mode = self._normalize_threading_mode(
            threading_mode=threading_mode
        )
        self._thread_dir = self._normalize_thread_dir(thread_dir=thread_dir)
        self._llm_adapter = llm_adapter
        if thread_name_rules is not None and len(thread_name_rules) > 0:
            self._thread_name_rules = [*thread_name_rules]
        else:
            self._thread_name_rules = [*DEFAULT_CHANNEL_THREAD_NAME_RULES]
        self._thread_list_document: MeshDocument | None = None
        self._thread_list_path: str | None = None

    @property
    def room(self) -> RoomClient:
        return self._room

    def thread_name_adapter(self) -> LLMAdapter | None:
        return self._llm_adapter

    @staticmethod
    def _normalize_threading_mode(*, threading_mode: str | None) -> str | None:
        if threading_mode is None:
            return None

        normalized = threading_mode.strip()
        if normalized == "" or normalized == "none":
            return None
        return normalized

    @staticmethod
    def _normalize_thread_dir(*, thread_dir: str | None) -> str | None:
        if thread_dir is None:
            return None

        normalized = thread_dir.strip().rstrip("/")
        if normalized == "":
            raise ValueError("thread_dir must not be empty")
        return normalized

    def _default_thread_dir_fallback_name(self) -> str:
        return "chat"

    def _default_thread_name(self) -> str:
        return "New Chat"

    def _uses_explicit_thread_dir_for_thread_list(self) -> bool:
        return False

    def _default_thread_dir(self) -> str:
        local_name = self._default_thread_dir_fallback_name()
        participant_name = self._room.local_participant.get_attribute("name")
        if isinstance(participant_name, str) and participant_name.strip() != "":
            local_name = participant_name.strip()

        return self._normalize_thread_dir(
            thread_dir=posixpath.join(".threads", local_name)
        )

    def _thread_list_dir(self) -> str | None:
        if (
            self._thread_dir is not None
            and self._uses_explicit_thread_dir_for_thread_list()
        ):
            return self._thread_dir
        if self._threading_mode == "default-new":
            return self._get_thread_dir()
        return None

    def _get_thread_dir(self) -> str:
        if self._thread_dir is not None:
            return self._thread_dir
        return self._default_thread_dir()

    def get_thread_dir(self) -> str:
        return self._get_thread_dir()

    @staticmethod
    def _sanitize_thread_name(*, value: str) -> str:
        normalized = value.strip()
        if normalized.endswith(".thread"):
            normalized = normalized[: -len(".thread")]

        normalized = re.sub(r"[-_/]+", " ", normalized)
        normalized = re.sub(r"\s+", " ", normalized).strip(" .-_")
        normalized = re.sub(r"[^A-Za-z0-9 .,!?':()&]+", "", normalized)
        normalized = re.sub(r"\s+", " ", normalized).strip(" .-_")
        if normalized == "":
            return "New Chat"
        if normalized == normalized.lower() or normalized == normalized.upper():
            normalized = normalized.title()
        return normalized[:64].strip() or "New Chat"

    @staticmethod
    def _thread_path_for_name(*, thread_name: str, thread_dir: str) -> str:
        return posixpath.join(thread_dir, f"{thread_name}.thread")

    async def _publish_thread_attributes(self) -> None:
        if self._threading_mode is not None:
            await self._room.local_participant.set_attribute(
                "meshagent.chatbot.threading",
                self._threading_mode,
            )
        thread_dir = self._thread_list_dir()
        if thread_dir is not None:
            await self._room.local_participant.set_attribute(
                "meshagent.chatbot.thread-dir",
                thread_dir,
            )
        thread_list_path = self._thread_list_index_path()
        if thread_list_path is not None:
            await self._room.local_participant.set_attribute(
                "meshagent.chatbot.thread-list",
                thread_list_path,
            )

    async def publish_thread_attributes(self) -> None:
        await self._publish_thread_attributes()

    def _utc_now_iso(self) -> str:
        return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

    def _thread_list_index_path(self) -> str | None:
        thread_dir = self._thread_list_dir()
        if thread_dir is None:
            return None
        return posixpath.join(thread_dir, "index.threadl")

    def _thread_list_entry_name_for_path(self, *, path: str) -> str:
        filename = posixpath.basename(path.strip())
        if filename.endswith(".thread"):
            filename = filename[: -len(".thread")]
        raw_name = filename.strip()
        if raw_name != "":
            try:
                parsed_uuid = uuid.UUID(raw_name)
            except ValueError:
                parsed_uuid = None
            if parsed_uuid is not None and str(parsed_uuid) == raw_name.lower():
                return self._default_thread_name()

        normalized = self._sanitize_thread_name(value=filename)
        if normalized == "New Chat":
            return self._default_thread_name()
        return normalized

    def _find_thread_list_entry(self, *, path: str) -> Element | None:
        if self._thread_list_document is None:
            return None

        for child in self._thread_list_document.root.get_children():
            if child.tag_name != "thread":
                continue
            if child.get_attribute("path") == path:
                return child

        return None

    def _is_index_managed_path(self, *, path: str) -> bool:
        thread_dir = self._thread_list_dir()
        if thread_dir is None:
            return False

        normalized_path = path.strip().strip("/")
        normalized_dir = thread_dir.strip().strip("/")
        if normalized_path == "" or normalized_dir == "":
            return False
        return normalized_path == normalized_dir or normalized_path.startswith(
            f"{normalized_dir}/"
        )

    def _upsert_thread_list_entry(
        self,
        *,
        path: str,
        name: str | None = None,
        created_at: str | None = None,
        modified_at: str | None = None,
    ) -> None:
        if self._thread_list_document is None or not self._is_index_managed_path(
            path=path
        ):
            return

        now = self._utc_now_iso()
        provided_name = name.strip() if isinstance(name, str) else ""
        entry = self._find_thread_list_entry(path=path)
        if entry is None:
            resolved_created_at = (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
            resolved_modified_at = (
                modified_at.strip()
                if isinstance(modified_at, str) and modified_at.strip() != ""
                else resolved_created_at
            )
            self._thread_list_document.root.append_child(
                tag_name="thread",
                attributes={
                    "name": (
                        provided_name
                        if provided_name != ""
                        else self._thread_list_entry_name_for_path(path=path)
                    ),
                    "path": path,
                    "created_at": resolved_created_at,
                    "modified_at": resolved_modified_at,
                },
            )
            return

        entry.set_attribute("path", path)
        if provided_name != "":
            entry.set_attribute("name", provided_name)
        else:
            existing_name = entry.get_attribute("name")
            if not isinstance(existing_name, str) or existing_name.strip() == "":
                entry.set_attribute(
                    "name",
                    self._thread_list_entry_name_for_path(path=path),
                )

        existing_created_at = entry.get_attribute("created_at")
        resolved_created_at = (
            existing_created_at.strip()
            if isinstance(existing_created_at, str)
            and existing_created_at.strip() != ""
            else (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
        )
        entry.set_attribute("created_at", resolved_created_at)
        resolved_modified_at = self._next_thread_list_modified_at(
            entry=entry,
            modified_at=modified_at,
        )
        entry.set_attribute("modified_at", resolved_modified_at)

    def bump_thread(self, *, path: str, name: str | None = None) -> None:
        entry = self._find_thread_list_entry(path=path)
        resolved_name = name
        if entry is not None:
            existing_name = entry.get_attribute("name")
            if isinstance(existing_name, str) and existing_name.strip() != "":
                resolved_name = None
        self._upsert_thread_list_entry(
            path=path,
            name=resolved_name,
            modified_at=self._utc_now_iso(),
        )

    def _next_thread_list_modified_at(
        self,
        *,
        entry: Element | None,
        modified_at: str | None = None,
    ) -> str:
        candidate = (
            modified_at.strip()
            if isinstance(modified_at, str) and modified_at.strip() != ""
            else self._utc_now_iso()
        )
        if entry is None:
            return candidate

        existing_modified_at = self._parse_iso_datetime(
            value=entry.get_attribute("modified_at")
        )
        candidate_modified_at = self._parse_iso_datetime(value=candidate)
        if (
            existing_modified_at is None
            or candidate_modified_at is None
            or candidate_modified_at > existing_modified_at
        ):
            return candidate

        return (
            (existing_modified_at + timedelta(microseconds=1))
            .isoformat()
            .replace(
                "+00:00",
                "Z",
            )
        )

    def _thread_list_entries(self) -> list[Element]:
        if self._thread_list_dir() is None or self._thread_list_document is None:
            return []

        return [
            child
            for child in self._thread_list_document.root.get_children()
            if child.tag_name == "thread"
        ]

    @staticmethod
    def _parse_iso_datetime(*, value: Any) -> datetime | None:
        if not isinstance(value, str):
            return None

        raw = value.strip()
        if raw == "":
            return None

        normalized = raw[:-1] + "+00:00" if raw.endswith("Z") else raw
        try:
            parsed = datetime.fromisoformat(normalized)
        except ValueError:
            return None

        if parsed.tzinfo is None:
            return parsed.replace(tzinfo=timezone.utc)
        return parsed

    def _thread_sort_datetime(self, *, entry: Element) -> datetime:
        modified_at = self._parse_iso_datetime(value=entry.get_attribute("modified_at"))
        if modified_at is not None:
            return modified_at
        created_at = self._parse_iso_datetime(value=entry.get_attribute("created_at"))
        if created_at is not None:
            return created_at
        return datetime.min.replace(tzinfo=timezone.utc)

    def _sorted_thread_list_entries(self) -> list[Element]:
        return sorted(
            self._thread_list_entries(),
            key=lambda entry: self._thread_sort_datetime(entry=entry),
            reverse=True,
        )

    @staticmethod
    def _thread_list_slice(
        *,
        entries: list[Element],
        limit: int,
        offset: int,
    ) -> list[Element]:
        normalized_offset = max(0, int(offset))
        normalized_limit = max(1, min(200, int(limit)))
        return entries[normalized_offset : normalized_offset + normalized_limit]

    async def _open_thread_list_document(self) -> None:
        index_path = self._thread_list_index_path()
        if index_path is None:
            return

        if (
            self._thread_list_document is not None
            and self._thread_list_path == index_path
        ):
            return

        self._thread_list_document = await self._room.sync.open(
            path=index_path,
            schema=thread_list_schema,
        )
        self._thread_list_path = index_path

    async def open_thread_list_document(self) -> None:
        await self._open_thread_list_document()

    async def _close_thread_list_document(self) -> None:
        thread_list_path = self._thread_list_path
        if self._thread_list_document is None or thread_list_path is None:
            return

        self._thread_list_document = None
        self._thread_list_path = None
        await self._room.sync.close(path=thread_list_path)

    async def close_thread_list_document(self) -> None:
        await self._close_thread_list_document()

    async def _next_available_thread_path(self, *, base_path: str) -> str:
        try:
            exists = await self._room.storage.exists(path=base_path)
        except Exception:
            return base_path

        if not exists:
            return base_path

        thread_dir, filename = posixpath.split(base_path)
        if filename.endswith(".thread"):
            base_name = filename[: -len(".thread")]
        else:
            base_name = filename

        for index in range(2, 1000):
            candidate = posixpath.join(thread_dir, f"{base_name} {index}.thread")
            try:
                if not await self._room.storage.exists(path=candidate):
                    return candidate
            except Exception:
                return candidate

        return posixpath.join(thread_dir, f"{base_name}-{uuid.uuid4().hex[:8]}.thread")

    @staticmethod
    def _attachment_name_for_thread_title(*, attachment: str) -> str:
        normalized = attachment.strip()
        if normalized == "":
            return ""

        parsed = urlparse(normalized)
        if parsed.scheme != "":
            attachment_name = PurePosixPath(parsed.path).name
            if attachment_name != "":
                return attachment_name

        attachment_name = PurePosixPath(normalized).name
        if attachment_name != "":
            return attachment_name

        return normalized

    def _normalized_thread_title_attachments(
        self,
        *,
        attachments: Sequence[str] | None,
    ) -> list[str]:
        attachment_names: list[str] = []
        for attachment in attachments or []:
            normalized = self._attachment_name_for_thread_title(attachment=attachment)
            if normalized != "":
                attachment_names.append(normalized)
        return attachment_names

    def _thread_name_input_text(
        self,
        *,
        message_text: str,
        attachments: Sequence[str] | None = None,
    ) -> str:
        parts: list[str] = []
        normalized_message_text = message_text.strip()
        if normalized_message_text != "":
            parts.append(f"Message:\n{normalized_message_text}")

        attachment_names = self._normalized_thread_title_attachments(
            attachments=attachments
        )
        if len(attachment_names) > 0:
            attachment_lines = "\n".join(
                f"- {attachment_name}" for attachment_name in attachment_names
            )
            parts.append(f"Attachments:\n{attachment_lines}")

        if len(parts) == 0:
            return self._default_thread_name()
        return "\n\n".join(parts)

    def _fallback_thread_name(
        self,
        *,
        message_text: str,
        attachments: Sequence[str] | None = None,
    ) -> str:
        normalized_message_text = message_text.strip()
        if normalized_message_text != "":
            return self._sanitize_thread_name(value=normalized_message_text)

        attachment_names = self._normalized_thread_title_attachments(
            attachments=attachments
        )
        if len(attachment_names) > 0:
            return self._sanitize_thread_name(
                value=", ".join(attachment_names[:3]),
            )

        return self._default_thread_name()

    def fallback_thread_name(
        self,
        *,
        message_text: str,
        attachments: Sequence[str] | None = None,
    ) -> str:
        return self._fallback_thread_name(
            message_text=message_text,
            attachments=attachments,
        )

    async def _determine_thread_name(
        self,
        *,
        message_text: str,
        attachments: Sequence[str] | None = None,
        on_behalf_of: Participant | None = None,
    ) -> str:
        generated_name = self._fallback_thread_name(
            message_text=message_text,
            attachments=attachments,
        )
        adapter = self.thread_name_adapter()
        if adapter is None:
            return generated_name

        session = adapter.create_session()
        cloned_context = session.copy()
        async with cloned_context:
            cloned_context.replace_rules(rules=self._thread_name_rules)
            cloned_context.append_user_message(
                self._thread_name_input_text(
                    message_text=message_text,
                    attachments=attachments,
                )
            )
            try:
                response = await adapter.next(
                    context=cloned_context,
                    room=self._room,
                    model=adapter.default_model(),
                    on_behalf_of=on_behalf_of,
                    toolkits=[],
                    output_schema={
                        "type": "object",
                        "required": ["thread_name"],
                        "additionalProperties": False,
                        "properties": {
                            "thread_name": {
                                "type": "string",
                                "description": "2-6 word topic name for the task thread",
                            }
                        },
                    },
                )
                if isinstance(response, dict):
                    thread_name = response.get("thread_name")
                    if isinstance(thread_name, str):
                        generated_name = self._sanitize_thread_name(
                            value=thread_name,
                        )
            except Exception as ex:
                logger.warning(
                    "unable to auto-generate thread name, using fallback",
                    exc_info=ex,
                )

        return generated_name

    async def _new_thread_path(self) -> str:
        base_path = self._thread_path_for_name(
            thread_name=str(uuid.uuid4()),
            thread_dir=self._get_thread_dir(),
        )
        return await self._next_available_thread_path(base_path=base_path)

    async def new_thread(
        self,
        *,
        message_text: str,
        attachments: Sequence[str] | None = None,
        on_behalf_of: Participant | None = None,
    ) -> tuple[str, str]:
        friendly_name = await self._determine_thread_name(
            message_text=message_text,
            attachments=attachments,
            on_behalf_of=on_behalf_of,
        )
        path = await self._new_thread_path()
        self.bump_thread(path=path, name=friendly_name)
        return path, friendly_name



import pytest

from meshagent.agents.responses_thread_adapter import (
    _headline_for_response_event,
    ResponsesThreadAdapter,
    _extract_image_dimensions,
    response_event_to_agent_event,
)


class _FakeParticipant:
    def __init__(self, *, name: str):
        self._name = name

    def get_attribute(self, key: str):
        if key == "name":
            return self._name
        return None


class _FakeRoom:
    def __init__(self):
        self.local_participant = _FakeParticipant(name="assistant")


def test_extract_image_dimensions_prefers_explicit_fields_then_size():
    width, height = _extract_image_dimensions(
        item={"width": 1536, "height": 1024},
        event={"size": "1024x1024"},
    )
    assert width == 1536
    assert height == 1024

    width, height = _extract_image_dimensions(
        item={"size": "512x768"},
        event=None,
    )
    assert width == 512
    assert height == 768


@pytest.mark.asyncio
async def test_emit_image_status_event_updates_image_element_state():
    adapter = object.__new__(ResponsesThreadAdapter)
    adapter._room = _FakeRoom()

    events: list[dict] = []
    writes: list[dict] = []

    async def _fake_handle_custom_event(*, messages, event):
        del messages
        events.append(event)

    def _fake_write_image(**kwargs):
        writes.append(kwargs)
        return kwargs.get("message_id", "")

    adapter.handle_custom_event = _fake_handle_custom_event  # type: ignore[assignment]
    adapter.write_image = _fake_write_image  # type: ignore[assignment]

    await adapter._emit_image_status_event(
        messages=object(),
        item_id="img-item-1",
        state="in_progress",
        headline="Generating image",
        width=1024,
        height=768,
    )

    assert len(events) == 1
    assert events[0]["state"] == "in_progress"
    assert events[0]["item_id"] == "img-item-1"

    assert len(writes) == 1
    assert writes[0]["message_id"] == "img-item-1"
    assert writes[0]["status"] == "generating"
    assert writes[0]["status_detail"] == "Generating image"
    assert writes[0]["width"] == 1024
    assert writes[0]["height"] == 768


def test_computer_call_events_are_not_classified_as_exec():
    event = {
        "type": "response.output_item.added",
        "item_id": "item_123",
        "item": {
            "id": "item_123",
            "type": "computer_call",
            "status": "in_progress",
        },
    }

    normalized = response_event_to_agent_event(event)
    assert isinstance(normalized, dict)
    assert normalized["kind"] == "tool"
    assert normalized["headline"] == "Using computer"


def test_computer_call_click_event_headline_is_friendly_without_coordinates():
    event = {
        "type": "response.output_item.added",
        "item_id": "item_456",
        "item": {
            "id": "item_456",
            "type": "computer_call",
            "status": "in_progress",
            "action": {
                "type": "click",
                "x": 140,
                "y": 320,
            },
        },
    }

    normalized = response_event_to_agent_event(event)
    assert isinstance(normalized, dict)
    assert normalized["kind"] == "tool"
    assert normalized["headline"] == "Clicking on page"
    assert normalized["details"] == []


def test_computer_call_scroll_event_uses_human_friendly_direction():
    event = {
        "type": "response.output_item.done",
        "item_id": "item_789",
        "item": {
            "id": "item_789",
            "type": "computer_call",
            "status": "completed",
            "action": {
                "type": "scroll",
                "scroll_x": 0,
                "scroll_y": -640,
            },
        },
    }

    normalized = response_event_to_agent_event(event)
    assert isinstance(normalized, dict)
    assert normalized["kind"] == "tool"
    assert normalized["headline"] == "Scrolled page"
    assert normalized["details"] == ["Direction: up"]


def test_computer_startup_event_uses_startup_specific_headline():
    event = {
        "type": "response.output_item.done",
        "item_id": "startup_1",
        "name": "computer.startup",
        "state": "completed",
    }

    headline = _headline_for_response_event(event=event, kind="tool", state="completed")
    assert headline == "Computer ready"



from typing import Any

import pytest

from meshagent.agents.adapter import (
    DEFAULT_MAX_TOOL_CALL_LENGTH,
    DEFAULT_MAX_TOOL_CALL_LINES,
    ToolResponseAdapter,
)
from meshagent.api.messaging import FileContent, JsonContent, TextContent


class _FakeRoom:
    pass


class _FakeToolResponseAdapter(ToolResponseAdapter):
    async def to_plain_text(self, *, room, response):
        del room
        del response
        return ""

    async def create_messages(
        self,
        *,
        context,
        tool_call: Any,
        room,
        response,
    ) -> list:
        del context
        del tool_call
        del room
        del response
        return []


def test_tool_response_adapter_defaults_match_opencode() -> None:
    adapter = _FakeToolResponseAdapter()

    assert adapter.max_tool_call_lines == DEFAULT_MAX_TOOL_CALL_LINES == 2000
    assert adapter.max_tool_call_length == DEFAULT_MAX_TOOL_CALL_LENGTH == 50 * 1024


def test_truncate_returns_text_content_for_long_text() -> None:
    adapter = _FakeToolResponseAdapter(max_tool_call_length=16, max_tool_call_lines=2)

    truncated = adapter.truncate(
        room=_FakeRoom(),
        content=TextContent(text="line1\nline2\nline3\nline4"),
    )

    assert isinstance(truncated, TextContent)
    assert "line1\nline2" in truncated.text
    assert "line3" not in truncated.text
    assert "The tool call returned too much data and was truncated." in truncated.text


def test_truncate_returns_text_content_for_long_json() -> None:
    adapter = _FakeToolResponseAdapter(max_tool_call_length=12, max_tool_call_lines=10)

    truncated = adapter.truncate(
        room=_FakeRoom(),
        content=JsonContent(json={"message": "x" * 50}),
    )

    assert isinstance(truncated, TextContent)
    assert "The tool call returned too much data and was truncated." in truncated.text


def test_truncate_leaves_file_content_unchanged() -> None:
    adapter = _FakeToolResponseAdapter(max_tool_call_length=8, max_tool_call_lines=1)
    content = FileContent(
        data=b"abcdefghijklmnopqrstuvwxyz",
        name="big.txt",
        mime_type="text/plain",
    )

    assert adapter.truncate(room=_FakeRoom(), content=content) is content


@pytest.mark.asyncio
async def test_file_content_to_text_content_decodes_declared_text_file() -> None:
    adapter = _FakeToolResponseAdapter()
    content = FileContent(
        data="hello\nworld".encode("utf-8"),
        name="notes.txt",
        mime_type="text/plain; charset=utf-8",
    )

    text_content = await adapter.file_content_to_text_content(
        room=_FakeRoom(),
        content=content,
    )

    assert isinstance(text_content, TextContent)
    assert text_content.text == "hello\nworld"


@pytest.mark.asyncio
async def test_file_content_to_text_content_decodes_utf8_text_without_text_mime() -> (
    None
):
    adapter = _FakeToolResponseAdapter()
    content = FileContent(
        data="plain utf8".encode("utf-8"),
        name="README",
        mime_type="application/octet-stream",
    )

    text_content = await adapter.file_content_to_text_content(
        room=_FakeRoom(),
        content=content,
    )

    assert isinstance(text_content, TextContent)
    assert text_content.text == "plain utf8"


@pytest.mark.asyncio
async def test_file_content_to_text_content_rejects_binary_utf8_file() -> None:
    adapter = _FakeToolResponseAdapter()
    content = FileContent(
        data=b"\x00\x01\x02",
        name="binary.bin",
        mime_type="application/octet-stream",
    )

    assert (
        await adapter.file_content_to_text_content(
            room=_FakeRoom(),
            content=content,
        )
        is None
    )


def test_tool_response_adapter_rejects_invalid_limits() -> None:
    with pytest.raises(ValueError):
        _FakeToolResponseAdapter(max_tool_call_length=0)

    with pytest.raises(ValueError):
        _FakeToolResponseAdapter(max_tool_call_lines=0)
